package main

import (
	"fmt"
	"log"
	"os"
	"strings"

	"git.pbiernat.dev/egommerce/basket-service/internal/app/config"
	"git.pbiernat.dev/egommerce/basket-service/internal/app/database"
	"git.pbiernat.dev/egommerce/basket-service/internal/app/server"
	"git.pbiernat.dev/egommerce/basket-service/internal/app/service"
	"git.pbiernat.dev/egommerce/basket-service/pkg/amqp"
	"git.pbiernat.dev/egommerce/basket-service/pkg/fluentd"
	"github.com/google/uuid"
)

const (
	defAppName       = "basket-worker"
	defLoggerAddr    = "api-logger:24224"
	defDbURL         = "postgres://postgres:12345678@postgres-db:5432/egommerce"
	defMongoDbURL    = "mongodb://mongodb:12345678@mongo-db:27017"
	defEventBusURL   = "amqp://guest:guest@api-eventbus:5672"
	ebEventsExchange = "api-events"
	ebEventsQueue    = "basket-worker"
)

func main() {
	if config.ErrLoadingEnvs != nil {
		log.Panicln("Error loading .env file", config.ErrLoadingEnvs)
	}

	id := uuid.New().String()[24:]
	c := new(server.Config)
	c.AppName = config.GetEnv("APP_NAME", defAppName) + "#:" + id
	c.LoggerAddr = config.GetEnv("LOGGER_ADDR", defLoggerAddr)
	c.DbURL = config.GetEnv("DATABASE_URL", defDbURL)
	c.EventBusURL = config.GetEnv("EVENTBUS_URL", defEventBusURL)
	c.EventBusExchange = ebEventsExchange
	c.EventBusQueue = ebEventsQueue
	// c.EventBusQueue = fmt.Sprintf("%s-%s", ebEventsQueue, id)

	logHost, logPort := fluentd.ParseAddr(c.LoggerAddr)
	logger := fluentd.NewLogger(c.AppName, logHost, logPort)
	defer logger.Close()

	// db conn
	dbConn, err := database.Connect(c.DbURL)
	if err != nil { // fixme: add wait-for-db...
		logger.Log("Failed to connect to Database server: %v\n", err)
		os.Exit(1)
	}
	defer dbConn.Close()

	// eventbus conn
	ebConn, ebCh, err := amqp.Open(c.EventBusURL)
	if err != nil {
		logger.Log("Failed to connect to EventBus server: %v\n", err)
		os.Exit(1)
	}
	defer ebCh.Close()
	defer amqp.Close(ebConn)

	err = amqp.NewExchange(ebCh, c.EventBusExchange)
	if err != nil {
		logger.Log("Failed to declare EventBus exchange: %v\n", err)
		os.Exit(1)
	}

	// create and bind queues
	_, err = ebCh.QueueDeclare(
		c.EventBusQueue, // name
		false,           // durable
		false,           // delete when unused
		false,           // exclusive
		false,           // no-wait
		nil,             // arguments
	)
	if err != nil {
		logger.Log("Failed to declare EventBus queue: %v\n", err)
		os.Exit(1)
	}

	amqp.BindQueueToExchange(ebCh, c.EventBusQueue, c.EventBusExchange, "catalog.basket.productAddedToBasket")
	amqp.BindQueueToExchange(ebCh, c.EventBusQueue, c.EventBusExchange, "catalog.basket.productRemovedFromBasket")
	amqp.BindQueueToExchange(ebCh, c.EventBusQueue, c.EventBusExchange, "catalog.basket.updateQuantity")
	// err = amqp.BindQueueToExchange(ebCh, c.EventBusQueue, c.EventBusExchange, "catalog-svc.product.addedToBasket")
	// if err != nil {
	// 	logger.Log("Failed to prepare EventBus queue: %v\n", err)
	// 	os.Exit(1)
	// }

	msgs, err := ebCh.Consume(
		c.EventBusQueue, // queue
		"",              // consumer
		false,           // auto-ack
		false,           // exclusive
		false,           // no-local
		false,           // no-wait
		nil,             // args
	)
	if err != nil {
		logger.Log("Failed to register a consumer: %s", err)
		os.Exit(1)
	}

	var forever chan struct{}
	go func() {
		basket := service.NewBasketService(dbConn, ebCh, logger)

		for d := range msgs {
			msg, err := amqp.Deserialize(d.Body)
			if err != nil {
				logger.Log("json error:", err)
				d.Reject(false) // FIXME: how to handle erros in queue...????
				continue
			}

			event := fmt.Sprintf("%s", msg["event"])
			data := (msg["data"]).(map[string]interface{})
			logger.Log("Message<%s>: %#v", event, data)

			basketID := data["basket_id"].(string) // FIXME CHECK basket_id!!!

			switch true {
			case strings.Contains(event, amqp.EVENT_PRODUCT_ADDED_TO_BASKET):
				// create new basket and add product to it
				// FIXME CHECK if basket exists
				_, err := basket.Get(basketID)
				if err != nil {
					logger.Log("Basket#%s not found. Creating...", basketID)
					err := basket.Create(basketID)
					if err != nil {
						logger.Log("Creating basket error: %s", err)
						d.Reject(false)
						continue
					}
				}

			case strings.Contains(event, amqp.EVENT_PRODUCT_REMOVED_FROM_BASKET):
				// remove product from basket
			}
			logger.Log("ACK: %s", event)
			d.Ack(false)
		}
	}()

	logger.Log("Waiting for messages...")
	<-forever
}

// create basket if not exists, next update product quantity
func updateBasketState() {

}
